library(dplyr)
library(future)

# globals ----------------------------------------------------------------------
start_ <- as.POSIXct("2013-03-31 00:00:00", tz = "UTC")
end_   <- as.POSIXct('2023-03-31 23:30:00', tz = "UTC")

years_unplanned <- 10
years_planned <- 7
mes_per_year <- 2.3

small_edb_invercargill <- 734  # 726 for DPP3
small_edb_nelson <- 327        # 329 for DPP3

saidi_collar_rateup <- 2
saifi_collar_rateup <- 2
saidi_cap_rateup <- 5
saifi_cap_rateup <- 5

# utils ------------------------------------------------------------------------
source("R/utils.R")

# import data ------------------------------------------------------------------
dpp <- data.table::fread("data-raw/edb_interruptions.csv.gz") |> data.frame()

# half hours -------------------------------------------------------------------
all_halfhours <- data.frame(time = seq(start_, end_, 60 * 30))

# boundaries -------------------------------------------------------------------
boundaries <- dpp |> select(edb) |> distinct() |>
  mutate(
    boundary = case_when(
      edb == "Electricity Invercargill" ~ small_edb_invercargill,
      edb == "Nelson Electricity" ~ small_edb_nelson,
      TRUE ~ mes_per_year * 48 * years_unplanned
    )
  )

# unplanned boundary values ----------------------------------------------------
# cut events for dates outside range
# create time variable rounded down to nearest half hour
unplanned <- dpp |>
  filter(planned_or_unplanned == "unplanned") |>
  filter(as.Date(start_date_and_time) >= as.Date(start_)) |>
  filter(as.Date(start_date_and_time) <= as.Date(end_)) |>
  mutate(
    time = round_down_to_nearest_half_hour(start_date_and_time)
  ) |>
  select(edb, time, saidi_value, saifi_value)

# create scaffold--all unique edb x half-hour combinations
edb_halfhours <- unplanned |>
  select(edb) |> distinct() |> cross_join(all_halfhours)

# merge unplanned with scaffold
unplanned_ <- left_join(edb_halfhours, unplanned, by = c("edb", "time")) |>
  arrange(edb, time) |>
  mutate(
    saidi_value = ifelse(is.na(saidi_value), 0, saidi_value),
    saifi_value = ifelse(is.na(saifi_value), 0, saifi_value)
  )

# get sum of saidi and saifi by half-hour
half_hour_totals <- unplanned_ |>
  group_by(edb, time) |>
  summarise(
    saidi = sum(saidi_value),
    saifi = sum(saifi_value),
    .groups = "drop"
  )

# get number of half-hour periods with a non-zero saidi value
half_hour_counts <- half_hour_totals |>
  group_by(edb) |>
  summarise(n = sum(saidi > 0))

# get rolling 24-hour totals
# cumsum(x)[t] - cumsum(x)[t -]
rolling_total <- half_hour_totals |>
  group_by(edb) |>
  mutate(
    saidi_roll = cumsum(saidi) - lag(cumsum(saidi), 48),
    saifi_roll = cumsum(saifi) - lag(cumsum(saifi), 48)
  ) |>
  ungroup()

# get saidi boundary values
boundary_saidi <-
  get_boundary(rolling_total, "edb", "saidi_roll", "saidi_bound", boundaries)

# get saifi boundary values
boundary_saifi <-
  get_boundary(rolling_total, "edb", "saifi_roll", "saifi_bound", boundaries)

# merge results
boundary_values <- half_hour_counts |>
  left_join(boundary_saidi, by = "edb") |>
  left_join(boundary_saifi, by = "edb")

## normalise -------------------------------------------------------------------
half_hour_cumsum <- rolling_total |>
  mutate(
    date = as.Date(time),
    year = year(time),
    month = month(time)
  ) |>
  mutate(
    disc_year = ifelse(month >= 4, year + 1, year)
  ) |>
  group_by(edb) |>
  mutate(
    cum_saidi = cumsum(saidi),
    cum_saifi = cumsum(saifi)
  ) |>
  left_join(
    boundary_values, by = "edb"
  )

# n.b. this takes a while.
for (i in 1:48) {
  half_hour_cumsum <- half_hour_cumsum |>
    group_by(edb) |>
    mutate(
      !! sprintf("me_saidi_%02d", i) :=
        ifelse(
          (mylead(cum_saidi, 48 - i + 1) - mylead(cum_saidi, 1 - i)) >= saidi_bound,
          1,
          0
        )
    ) |>
    mutate(
      saidi_major = rowSums(pick(starts_with("me_saidi_")), na.rm = TRUE)
    ) |>
    mutate(
      saidi_norm = case_when(
        (saidi_major > 0) & saidi > ((saidi_bound / 48)) ~ saidi_bound / 48,
        TRUE ~ saidi
      )
    )
}

# n.b. this takes a while.
for (i in 1L:48L) {
  half_hour_cumsum <- half_hour_cumsum |>
    group_by(edb) |>
    mutate(
      !! sprintf("me_saifi_%02d", i) :=
        ifelse(
          (mylead(cum_saifi, 48 - i + 1) - mylead(cum_saifi, 1 - i)) >= saifi_bound,
          1,
          0
        )
    ) |>
    mutate(
      saifi_major = rowSums(pick(starts_with("me_saifi_")), na.rm = TRUE)
    ) |>
    mutate(
      saifi_norm = case_when(
        saifi_major > 0 & saifi > saifi_bound / 48 ~ saifi_bound / 48,
        TRUE ~ saifi
      )
    )
}

# keep only last `years_unplanned` full years
yrs <- half_hour_cumsum$disc_year |>
  as.numeric() |>
  unique() |>
  sort() |>
  tail(years_unplanned)

half_hour_cumsum <- half_hour_cumsum |> filter(disc_year >= min(yrs))

# average for period of interest
targets <- half_hour_cumsum |>
  group_by(edb) |>
  summarise(
    saidi_target = sum(saidi_norm) / years_unplanned,
    saifi_target = sum(saifi_norm) / years_unplanned,
    .groups = "drop"
  )

# standard errors
se <- half_hour_cumsum |>
  group_by(edb, date) |>
  summarise(
    saidi_daily = sum(saidi_norm),
    saifi_daily = sum(saifi_norm),
    .groups = "drop"
  ) |> group_by(edb) |>
  summarise(
    saidi_sd = sd(saidi_daily) * sqrt(365.25),
    saifi_sd = sd(saifi_daily) * sqrt(365.25),
    .groups = "drop"
  )

targets <- targets |>
  inner_join(se, by = "edb") |>
  mutate(
    saidi_cap = saidi_target + 2 * saidi_sd,
    saifi_cap = saifi_target + 2 * saifi_sd
  )

## final unplanned parameters --------------------------------------------------
unplanned_parameters <- boundary_values |>
  inner_join(targets, by = "edb") |>
  mutate(saidi_collar = 0, saifi_collar = 0) |>
  select(
    edb, n,
    saidi_bound, saidi_target, saidi_collar, saidi_sd, saidi_cap,
    saifi_bound, saifi_target, saifi_collar, saifi_sd, saifi_cap
  )

write.csv(
  unplanned_parameters,
  "outputs/dpp4_draft_decision_unplanned_parameters_normalised_unadjusted_for_caps.csv",
  row.names = FALSE
)

# planned parameters -----------------------------------------------------------
planned <- dpp |>
  filter(planned_or_unplanned == "planned") |>
  filter(disclosure_year >= year(start_)) |>
  filter(start_date_and_time >= round_up_to_nearest_month(start_)) |>
  filter(start_date_and_time <= end_)

years <- planned$disclosure_year |> unique() |> sort() |> tail(years_planned)

planned_parameters <- planned |>
  filter(disclosure_year %in% years) |>
  group_by(edb) |>
  summarise(
    saidi_target = sum(saidi_value, na.rm = TRUE) / years_planned,
    saifi_target = sum(saifi_value, na.rm = TRUE) / years_planned,
    .groups = "drop"
  ) |>
  mutate(
    saidi_collar = 0,
    saifi_collar = 0,
    saidi_cap = saidi_collar_rateup * saidi_target,
    saifi_cap = saifi_collar_rateup * saifi_target,
    saidi_limit = saidi_cap_rateup * saidi_cap,
    saifi_limit = saifi_cap_rateup * saifi_cap
  ) |>
  select(
    edb,
    saidi_collar, saidi_target, saidi_cap, saidi_limit,
    saifi_collar, saifi_target, saifi_cap, saifi_limit
  )

write.csv(
  planned_parameters,
  "outputs/dpp4_draft_decision_planned_parameters_unadjusted_for_caps.csv",
  row.names = FALSE
)
